//
//  MoonPhaseGLView.m
//  MoonPhase
//
//  Created by James Dessart on 04/09/08.
//  Copyright 2008 __MyCompanyName__. All rights reserved.
//

#import "MoonPhaseGLView.h"

#include "phase_of_the_moon.h"

#define TWOPI	(2 * M_PI)
#define PID2	(M_PI / 2)
#define RADIAN_FROM_PHASE(x) (TWOPI * (x * (1.0/MAX_PHASES)))

static NSString *moonPhaseStrings[] =
{
	@"New Moon",
	@"Waxing Crescent Moon",
	@"First Quarter Moon",
	@"Waxing Gibbous Moon",
	@"Full Moon",
	@"Waning Gibbous Moon",
	@"Last Quarter Moon",
	@"Waning Crescent Moon",
	0
};

/*
   Create a sphere centered at c, with radius r, and precision n
   Draw a point for zero radius spheres
   From: http://local.wasp.uwa.edu.au/~pbourke/texture_colour/spheremap/
   Modified to use flat texture mapping
*/
typedef struct
{
	float x;
	float y;
	float z;
} XYZ;

void CreateSphere(MoonPhaseGLView *view,XYZ c,double r,int n)
{
	int i,j;
	double theta1,theta2,theta3;
	XYZ e,p,t;

	if (r < 0)
		r = -r;
	if (n < 0)
		n = -n;

	if (n < 4 || r <= 0)
	{
		glBegin(GL_POINTS);
		glVertex3f(c.x,c.y,c.z);
		glEnd();
		return;
	}

	for (j=0;j<n/2;j++)
	{
		theta1 = j * TWOPI / n - PID2;
		theta2 = (j + 1) * TWOPI / n - PID2;

		glBegin(GL_QUAD_STRIP);
		for (i=0;i<=n;i++)
		{
			theta3 = i * TWOPI / n;

			e.x = cos(theta2) * cos(theta3);
			e.y = sin(theta2);
			e.z = cos(theta2) * sin(theta3);
			p.x = c.x + r * e.x;
			p.y = c.y + r * e.y;
			p.z = c.z + r * e.z;

			t.x = ((e.x + 1.0) / 2.0) * view->textureSize.width;
			t.y = ((e.y - 1.0) / -2.0) * view->textureSize.height;

			//glTexCoord2f(i/(double)n,2*(j+1)/(double)n);
			glTexCoord2f(t.x, t.y);

			glNormal3f(e.x,e.y,e.z);
			glVertex3f(p.x,p.y,p.z);

			e.x = cos(theta1) * cos(theta3);
			e.y = sin(theta1);
			e.z = cos(theta1) * sin(theta3);
			p.x = c.x + r * e.x;
			p.y = c.y + r * e.y;
			p.z = c.z + r * e.z;

			t.x = ((e.x + 1.0) / 2.0) * view->textureSize.width;
			t.y = ((e.y - 1.0) / -2.0) * view->textureSize.height;

			//glTexCoord2f(i/(double)n,2*(j+1)/(double)n);
			glTexCoord2f(t.x, t.y);

			glNormal3f(e.x,e.y,e.z);
			glVertex3f(p.x,p.y,p.z);
		}
		glEnd();
	}
}

@implementation MoonPhaseGLView

- (id)initWithFrame:(NSRect)frameRect pixelFormat:(NSOpenGLPixelFormat *)format
{
	self = [super initWithFrame:frameRect pixelFormat:format];
	
	if( self != 0 )
	{
		phase = 0;
		textureName = 0;
	}
	
	return self;
}

- (void)textureFromImage:(NSImage*)theImg textureName:(GLuint*)texName
{
    NSBitmapImageRep* bitmap = [NSBitmapImageRep alloc];
    int samplesPerPixel = 0;
    NSSize imgSize = textureSize = [theImg size];
 
    [theImg lockFocus];
    [bitmap initWithFocusedViewRect:
                    NSMakeRect(0.0, 0.0, imgSize.width, imgSize.height)];
    [theImg unlockFocus];
 
    // Set proper unpacking row length for bitmap.
    glPixelStorei(GL_UNPACK_ROW_LENGTH, [bitmap pixelsWide]);
 
    // Set byte aligned unpacking (needed for 3 byte per pixel bitmaps).
    glPixelStorei (GL_UNPACK_ALIGNMENT, 1);
 
    // Generate a new texture name if one was not provided.
    if (*texName == 0)
        glGenTextures (1, texName);
    glBindTexture (GL_TEXTURE_RECTANGLE_EXT, *texName);
 
    // Non-mipmap filtering (redundant for texture_rectangle).
    glTexParameteri(GL_TEXTURE_RECTANGLE_EXT, GL_TEXTURE_MIN_FILTER,  GL_LINEAR);
    samplesPerPixel = [bitmap samplesPerPixel];
 
    // Nonplanar, RGB 24 bit bitmap, or RGBA 32 bit bitmap.
    if(![bitmap isPlanar] &&
        (samplesPerPixel == 3 || samplesPerPixel == 4))
    {
        glTexImage2D(GL_TEXTURE_RECTANGLE_EXT, 0,
            samplesPerPixel == 4 ? GL_RGBA8 : GL_RGB8,
            [bitmap pixelsWide],
            [bitmap pixelsHigh],
            0,
            samplesPerPixel == 4 ? GL_RGBA : GL_RGB,
            GL_UNSIGNED_BYTE,
            [bitmap bitmapData]);
    }
    else
    {
        // Handle other bitmap formats.
    }
 
    // Clean up.
    [bitmap release];
}

- (void)awakeFromNib
{
	[[self openGLContext] makeCurrentContext];
	
	glClearColor( 0.0f, 0.0f, 0.0f, 0.0f );

	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glEnable(GL_DEPTH_TEST);
	
	GLfloat emission[] = {0.3, 0.3, 0.3, 1.0};
	
	glMaterialfv(GL_FRONT, GL_EMISSION, emission);
	
	glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );
	
	[self textureFromImage:[NSImage imageNamed:@"moon.jpg"] textureName:&textureName];
	
	glEnable(GL_TEXTURE_RECTANGLE_EXT);

	[self setPhase:phase_of_the_moon()];
	[self setCurrentPhase:phase_of_the_moon()];
	
	// Schedule a timer to go off every hour (60 secs * 60 mins)
	[NSTimer scheduledTimerWithTimeInterval:3600 target:self
								   selector:@selector(timerTick:)
								   userInfo:self repeats:YES];
}

- (void)setPhase:(int)aPhase
{
	if( phase != aPhase )
	{
		phase = aPhase;
		[self setNeedsDisplay:YES];
		[self willChangeValueForKey:@"displayPhaseString"];
		[self didChangeValueForKey:@"displayPhaseString"];
	}
}

- (void)setCurrentPhase:(int)aPhase
{
	if( currentPhase != aPhase )
	{
		currentPhase = aPhase;
		[self willChangeValueForKey:@"currentPhaseString"];
		[self didChangeValueForKey:@"currentPhaseString"];
	}
}

- (NSString*)displayPhaseString
{
	return moonPhaseStrings[phase>>1];
}

- (NSString*)currentPhaseString
{
	return moonPhaseStrings[currentPhase>>1];
}

- (void)timerTick:(NSTimer*)theTimer
{
	[self setCurrentPhase:phase_of_the_moon()];
}

- (void)drawRect:(NSRect)rect
{
    // Drawing code here.
	GLfloat lightPosition[] = { 0, 0, 0, 0 };
	
	//int phase = phase_of_the_moon();
	
	double radiansForPhase = RADIAN_FROM_PHASE(phase);
	
	lightPosition[0] = sin(radiansForPhase);
	lightPosition[2] = -cos(radiansForPhase);

	glLightfv( GL_LIGHT0, GL_POSITION, lightPosition );

	glViewport( 0, 0, (GLsizei) rect.size.width, (GLsizei) rect.size.height );

	glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	
	glMatrixMode(GL_PROJECTION);
	
	glLoadIdentity();
	
	glOrtho( 0.0, (GLdouble) rect.size.width, 0.0, (GLdouble) rect.size.height,
			 -1000.0, 1000.0 );
	
	glMatrixMode(GL_MODELVIEW);
	
	glLoadIdentity();
	
	int smallerWidth =
		(rect.size.width > rect.size.height) ? rect.size.height : rect.size.width;
	GLdouble radius = (GLdouble)smallerWidth / 2 - (GLdouble)smallerWidth / 10;
	//GLdouble radius = (GLdouble)smallerWidth / 2;
	
	glTranslated( ((GLdouble) rect.size.width) / 2, ((GLdouble) rect.size.height) / 2, 0.0 );

#define CREATE_SPHERE 1

#if CREATE_SPHERE
	XYZ center = { 0, 0, 0 };
	
	CreateSphere( self, center, radius, 50 );
#endif

#if DRAW_QUADRIC
	GLUquadricObj *quadric = gluNewQuadric();
	
	gluQuadricDrawStyle( quadric, GLU_FILL );
	
	gluQuadricNormals( quadric, GLU_SMOOTH );
	
	gluQuadricTexture( quadric, GL_TRUE );
	
	gluSphere( quadric, radius, 50, 50 );
	
	gluDeleteQuadric(quadric);
#endif
	
#if DRAW_GLUT
	glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
	glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
	glEnable(GL_TEXTURE_GEN_S);
	glEnable(GL_TEXTURE_GEN_T);

	glutSolidSphere( radius, 50, 50 );
#endif
	
	[[self openGLContext] flushBuffer];
}

@end
